@@ -165,9 +165,11 @@ class Campaign:
 
     @osm.setter
     def osm(self, value):
-        for f in tqdm(self, desc="OSM Setter (Files in Campaign)"):
+        for f in self:
             f.osm = value
-            if self.map_matching:
+
+        if self.map_matching:
+            for f in tqdm(self, desc="Map Matching"):
                 f.do_map_matching()
 
         self._osm = value
@@ -425,7 +427,7 @@ class Campaign:
                 for f in files:
                     self.files.append(f)
         else:
-            for p in tqdm(file_paths, desc="file paths"):
+            for p in tqdm(file_paths, desc="File Import"):
                 self.files.append(RDYFile(path=p,
                                           sync_method=sync_method,
                                           timedelta_unit=timedelta_unit,
