@@ -62,8 +62,40 @@ class TimeSeries(ABC):
                                                                                   seconds=self.get_duration())),
                                                                               self.get_sample_rate())
 
-    def cutoff(self, timestamp_when_started: int, timestamp_when_stopped: int):
-        """ Cuts off measurement values saved before/after the measurement was started/stopped
+    def cut(self, start: float = 0, end: float = 0):
+        """ Cuts off seconds after start and before end
+
+        Parameters
+        ----------
+        start: float
+            Seconds to cutoff after start
+        end: float
+            Seconds to cutoff before end
+        """
+        if (start + end) >= self.get_duration():
+            raise ValueError(f'Trying to cut off more seconds than duration of {self.__class__.__name__}')
+
+        if len(self.time) > 0 and self.time[0] != 0:
+            d = self.__dict__.copy()
+
+            for key in ["filename", "rdy_format_version"]:
+                d.pop(key)
+
+            t = d["time"]
+            t_sec = (t - t[0]) / np.timedelta64(1, "s")
+
+            idxs = np.where(np.logical_and(t_sec >= start, t_sec <= (t_sec[-1] - end)))
+            for k, v in d.items():
+                if len(t) == len(v):
+                    self.__setattr__(k, v[idxs])
+
+            self._timedelta: np.ndarray = np.diff(self._time)
+        else:
+            logger.debug("(%s) Cannot cut %s if timeseries is empty or series already starts at 0" %
+                         (self.filename, self.__class__.__name__))
+
+    def trim_ends(self, timestamp_when_started: int, timestamp_when_stopped: int):
+        """ Trims measurement values saved before/after the measurement was started/stopped
 
         Parameters
         ----------
